from __future__ import absolute_import, division, print_function

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

import typing

import pytest
import six
import typing_extensions

import tippo

T = tippo.TypeVar("T")


def test_all():
    assert tippo._all_ is getattr(tippo, "__all__")

    assert set(typing.__all__).issubset(tippo._all_)  # type: ignore
    assert set(typing_extensions.__all__).issubset(tippo._all_)  # type: ignore

    for member in tippo._all_:
        assert not member.startswith("_")
        assert hasattr(tippo, member)

    if "GenericMeta" in typing.__all__:  # type: ignore
        assert "GenericMeta" in tippo._all_

    for _b, (names, _v) in tippo._GENERIC_TYPES.items():
        for name in names:
            assert name in tippo._all_

    assert "final" in tippo._all_
    assert "TypeAlias" in tippo._all_
    assert "ClassVar" in tippo._all_
    assert "get_origin" in tippo._all_
    assert "get_args" in tippo._all_
    assert "get_name" in tippo._all_


def test_missing():
    obj = object()
    for missing in ("TypeAlias", "ClassVar"):
        if not hasattr(typing, missing) and not hasattr(typing_extensions, missing):
            assert getattr(tippo, missing)[obj] is obj


def test_new_type():
    if not hasattr(typing, "NewType") and not hasattr(typing_extensions, "NewType"):
        obj = object()
        assert getattr(tippo, "NewType")("obj", obj) is obj


def test_generic_meta():

    # Metaclass uniformity.
    assert isinstance(tippo.Generic, tippo.GenericMeta)

    # Comparisons.
    assert tippo.Mapping[str, int] == tippo.Mapping[str, int]
    assert not (tippo.Mapping[str, int] != tippo.Mapping[str, int])

    # Bracket syntax should not work if class doesn't inherit from Generic.
    class Class(six.with_metaclass(tippo.GenericMeta, object)):
        pass

    with pytest.raises(TypeError):
        _BadClass = Class[None]  # type: ignore
        assert not _BadClass

    # Should error if didn't specify type variables to Generic.
    with pytest.raises(TypeError):

        class _BadClass(Class, tippo.Generic[3]):  # type: ignore
            pass

        assert not _BadClass

    with pytest.raises(TypeError):

        class _BadClass(Class, tippo.Generic[int]):  # type: ignore
            pass

        assert not _BadClass

    # Declare a proper Generic class and check inheritance, equality, and instantiation.
    class _Class(Class, tippo.Generic[T]):
        pass

    assert _Class[int]
    assert (_Class[int] == _Class[int]) is True
    assert (_Class[int] == _Class[(int,)]) is True
    assert (_Class[int] != _Class[(int,)]) is False

    assert isinstance(_Class[int](), _Class)
    assert isinstance(_Class[(int,)](), _Class)

    assert issubclass(_Class, Class)
    assert _Class[int]
    assert (_Class[int] == _Class[int]) is True
    assert (_Class[int] == _Class[(int,)]) is True
    assert (_Class[int] != _Class[(int,)]) is False

    assert isinstance(_Class[int](), _Class)
    assert isinstance(
        _Class[(int,)](),
        _Class,
    )

    # Test weakref slot with generic.
    class Class(tippo.Generic[T]):
        __slots__ = ("__weakref__",)

    class SubClass(Class[T]):
        __slots__ = ()

    assert SubClass


def test_generic_aliases():
    for original_base, info in tippo._GENERIC_TYPES.items():
        for name in info.names:
            generic = getattr(tippo, name)
            if original_base is not generic:
                assert not hasattr(original_base, "__class_getitem__")
                assert not hasattr(type(original_base), "__getitem__")
                assert hasattr(generic, "__class_getitem__") or hasattr(type(generic), "__getitem__")


def test_get_origin():
    assert tippo.get_origin(tippo.Literal[42]) is tippo.Literal
    assert tippo.get_origin(int) is None
    assert tippo.get_origin(tippo.ClassVar[int]) is tippo.ClassVar
    assert tippo.get_origin(tippo.Generic) is tippo.Generic
    assert tippo.get_origin(tippo.Generic[T]) is tippo.Generic
    assert tippo.get_origin(tippo.Union[T, int]) is tippo.Union
    assert tippo.get_origin(tippo.List[tippo.Tuple[T, T]][int]) in (list, tippo.List)

    assert tippo.get_origin(tippo.Callable) in (tippo.Callable, collections_abc.Callable)
    assert tippo.get_origin(tippo.Tuple) in (tippo.Tuple, tuple)
    assert tippo.get_origin(tippo.Dict) in (tippo.Dict, dict)
    assert tippo.get_origin(tippo.List) in (tippo.List, list)
    assert tippo.get_origin(tippo.Set) in (tippo.Set, set)
    assert tippo.get_origin(tippo.Union) is None
    assert tippo.get_origin(tippo.Literal) is None
    assert tippo.get_origin(tippo.Final) is None
    assert tippo.get_origin(tippo.ClassVar) is None

    assert tippo.get_origin(tippo.Literal[42]) is tippo.Literal
    assert tippo.get_origin(int) is None
    assert tippo.get_origin(None) is None
    assert tippo.get_origin(tippo.ClassVar[int]) is tippo.ClassVar
    assert tippo.get_origin(tippo.Generic) is tippo.Generic
    assert tippo.get_origin(tippo.Generic[T]) is tippo.Generic
    assert tippo.get_origin(tippo.Union[T, int]) is tippo.Union
    assert tippo.get_origin(tippo.List[tippo.Tuple[T, T]][int]) in (tippo.List, list)
    assert tippo.get_origin(tippo.Callable[[str, int], bool]) in (tippo.Callable, collections_abc.Callable)
    assert tippo.get_origin(tippo.Tuple[str]) in (tippo.Tuple, tuple)
    assert tippo.get_origin(tippo.Dict[str, int]) in (tippo.Dict, dict)
    assert tippo.get_origin(tippo.List[str]) in (tippo.List, list)
    assert tippo.get_origin(tippo.Set[str]) in (tippo.Set, set)
    assert tippo.get_origin(tippo.Union[str, int]) is tippo.Union
    assert tippo.get_origin(tippo.Literal[True, 0]) is tippo.Literal
    assert tippo.get_origin(tippo.Final[int]) is tippo.Final
    assert tippo.get_origin(tippo.ClassVar[int]) is tippo.ClassVar


def test_get_args():
    assert tippo.get_args(tippo.Dict[str, int]) == (str, int)
    assert tippo.get_args(int) == ()
    assert tippo.get_args(tippo.Union[int, tippo.Union[T, int], str][int]) == (int, str)
    assert tippo.get_args(tippo.Union[int, tippo.Tuple[T, int]][str]) == (int, tippo.Tuple[str, int])
    assert tippo.get_args(tippo.Callable[[], T][int]) == ([], int)

    assert tippo.get_args(tippo.Dict) == ()
    assert tippo.get_args(int) == ()
    assert tippo.get_args(None) == ()
    assert tippo.get_args(tippo.Union) == ()
    assert tippo.get_args(tippo.Callable) == ()

    assert tippo.get_args(tippo.Dict[str, int]) == (str, int)
    assert tippo.get_args(tippo.Union[int, tippo.Union[T, int], str][int]) == (int, str)
    assert tippo.get_args(tippo.Union[int, tippo.Tuple[T, int]][str]) == (int, tippo.Tuple[str, int])
    assert tippo.get_args(tippo.Callable[[float, bool], T][int]) == ([float, bool], int)
    assert tippo.get_args(tippo.Callable[[], T][int]) == ([], int)
    assert tippo.get_args(tippo.Callable[Ellipsis, T][int]) == (Ellipsis, int)

    assert tippo.get_args(tippo.Union[int, tippo.Tuple[T, tippo.Dict[str, bool]]][bool]) == (
        int,
        tippo.Tuple[bool, tippo.Dict[str, bool]],
    )
    assert tippo.get_args(tippo.Dict[str, T][bool]) == (str, bool)
    assert tippo.get_args(tippo.List[T][bool]) == (bool,)
    assert tippo.get_args(tippo.Set[T][bool]) == (bool,)

    assert tippo.get_args(tippo.Union[int, tippo.Union[T, int], str]) == (int, T, str)
    assert tippo.get_args(tippo.Union[int, tippo.Tuple[T, int]]) == (int, tippo.Tuple[T, int])
    assert tippo.get_args(tippo.Callable[[float, bool], T]) == ([float, bool], T)
    assert tippo.get_args(tippo.Callable[[], T]) == ([], T)
    assert tippo.get_args(tippo.Callable[Ellipsis, T]) == (Ellipsis, T)


def test_get_name():
    assert tippo.get_name(None) == "None"
    assert tippo.get_name(Ellipsis) == "..."
    assert tippo.get_name(None) == "None"
    assert tippo.get_name(type(None)) == "NoneType"
    assert tippo.get_name(int) == "int"
    assert tippo.get_name(True) == "True"
    assert tippo.get_name(False) == "False"
    assert tippo.get_name(NotImplemented) == "NotImplemented"
    assert tippo.get_name(type(NotImplemented)) == "NotImplementedType"
    assert tippo.get_name(tippo.Any) == "Any"
    assert tippo.get_name(tippo.Optional) == "Optional"
    assert tippo.get_name(tippo.Generic) == "Generic"
    assert tippo.get_name(tippo.Literal) == "Literal"
    assert tippo.get_name(tippo.Final) == "Final"
    assert tippo.get_name(tippo.ForwardRef) == "ForwardRef"
    assert tippo.get_name(tippo.ClassVar) == "ClassVar"
    assert tippo.get_name(tippo.Callable) == "Callable"
    assert tippo.get_name(tippo.Tuple) == "Tuple"
    assert tippo.get_name(tippo.Union) == "Union"
    assert tippo.get_name(tippo.Mapping) == "Mapping"
    assert tippo.get_name(tippo.Union) == "Union"
    assert tippo.get_name(tippo.List) == "List"
    assert tippo.get_name(tippo.Dict) == "Dict"
    assert tippo.get_name(tippo.Set) == "Set"

    assert tippo.get_name(tippo.ForwardRef("Foo")) == "Foo"
    assert tippo.get_name(tippo.Generic[T]) == "Generic"
    assert tippo.get_name(tippo.Literal[True, False, "abc"]) == "Literal"
    assert tippo.get_name(tippo.Final[int]) == "Final"
    assert tippo.get_name(tippo.ClassVar[int]) == "ClassVar"
    assert tippo.get_name(tippo.Callable[[str, int], bool]) == "Callable"
    assert tippo.get_name(tippo.Callable[..., bool]) == "Callable"
    assert tippo.get_name(tippo.Tuple[str, int]) == "Tuple"
    assert tippo.get_name(tippo.Tuple[str, ...]) == "Tuple"
    assert tippo.get_name(tippo.Union[str, int]) == "Union"
    assert tippo.get_name(tippo.Mapping[str, int]) == "Mapping"
    assert tippo.get_name(tippo.Union[str, int]) == "Union"
    assert tippo.get_name(tippo.List[str]) == "List"
    assert tippo.get_name(tippo.Dict[str, int]) == "Dict"
    assert tippo.get_name(tippo.Set[str]) == "Set"
    assert tippo.get_name(tippo.Set["Foo"]) == "Set"
    assert tippo.get_name(tippo.List["Tuple"]) == "List"

    assert tippo.get_name(object()) is None


if __name__ == "__main__":
    pytest.main()



import functools as _functools
import types as _types
from weakref import ref  # noqa
from weakref import ReferenceType, WeakKeyDictionary, WeakSet, WeakValueDictionary

import six as _six

try:
    import collections.abc as _collections_abc
except ImportError:
    import collections as _collections_abc  # type: ignore

# Import typing.
import typing as _typing
from typing import *  # noqa

# Import typing extensions.
import typing_extensions as _typing_extensions
from typing_extensions import *  # type: ignore

# Prepare __all__ by combining typing + typing_extensions.
_all_ = _typing.__all__ + _typing_extensions.__all__  # type: ignore
globals()["__all__"] = _all_


def _update_all(*_members):
    _all_.extend(set(_members).difference(_all_))


# Forward reference type.
try:
    from typing import ForwardRef
except ImportError:
    from typing import _ForwardRef as ForwardRef  # type: ignore  # noqa

    _update_all("ForwardRef")


# Mappings.
_BUILTINS_MAPPING = {
    List: list,
    Set: set,
    Dict: dict,
    Tuple: tuple,
    Type: type,
}
_BUILTINS_MAPPING.update(
    dict(
        (getattr(_typing, n, getattr(_typing_extensions, n, None)), getattr(_collections_abc, n))
        for n in set(_collections_abc.__all__).intersection(
            _typing.__all__ + _typing_extensions.__all__  # type: ignore
        )
        if not n.startswith("_")
    )
)
assert None not in _BUILTINS_MAPPING
_TYPING_MAPPING = dict((b, t) for t, b in _BUILTINS_MAPPING.items())


def get_builtin(typ):
    """
    Get equivalent builtin.

    :param typ: Type.
    :return: Builtin type.
    """
    return _BUILTINS_MAPPING.get(typ, typ)


def get_typing(typ):
    """
    Get equivalent typing.

    :param typ: Type.
    :return: Typing type.
    """
    return _TYPING_MAPPING.get(typ, typ)


_update_all("get_builtin", "get_typing")


# Patch GenericMeta for Python 2.7 with some fixes.
try:
    from typing import GenericMeta as _GenericMeta  # type: ignore
except ImportError:
    GenericMeta = type
else:

    # Fix not equal operator logic in python 2.
    def __ne__(cls, other):
        is_equal = cls == other
        if is_equal is NotImplemented:
            return NotImplemented
        else:
            return not is_equal

    __ne__.__module__ = _GenericMeta.__module__
    type.__setattr__(_GenericMeta, "__ne__", __ne__)

    # Fix subclassing slotted generic class with __weakref__.
    _original_getitem = getattr(_GenericMeta, "__getitem__")

    @_functools.wraps(_original_getitem)
    def __getitem__(cls, params):
        slots = getattr(cls, "__slots__", None)
        if slots is not None and "__weakref__" in slots:
            type.__setattr__(cls, "__slots__", tuple(s for s in slots if s != "__weakref__"))
            try:
                return _original_getitem(cls, params)  # type: ignore
            finally:
                type.__setattr__(cls, "__slots__", slots)
        else:
            return _original_getitem(cls, params)  # type: ignore

    type.__setattr__(_GenericMeta, "__getitem__", __getitem__)

_update_all("GenericMeta")


# Add missing final decorator for older Python versions.
if "final" not in globals():

    def final(f):  # type: ignore
        """A decorator to indicate final methods and final classes."""
        return f

    _update_all("final")


class _MissingMeta(type):
    def __getitem__(cls, _):
        return cls


# Add missing TypeAlias for older Python versions.
if "TypeAlias" not in globals():

    class TypeAlias(_six.with_metaclass(_MissingMeta, object)):  # type: ignore
        pass

    _update_all("TypeAlias")


# Add missing ClassVar for older Python versions.
if "ClassVar" not in globals():

    class ClassVar(_six.with_metaclass(_MissingMeta, object)):  # type: ignore
        pass

    _update_all("ClassVar")


# Add missing NewType for older Python versions.
if "NewType" not in globals():

    def NewType(_name, _typ):  # type: ignore
        return _typ

    _update_all("NewType")


# Add missing inspection functions for older Python versions.
if "get_origin" not in globals():
    from typing_inspect import get_origin as _get_origin  # type: ignore

    @_functools.wraps(_get_origin)
    def get_origin(typ):
        if typ is Generic:
            return Generic

        if typ in (Union, Literal, Final, ClassVar):
            return None

        for name, origin in {"_Literal": Literal, "_ClassVar": ClassVar, "_Final": Final}.items():
            if type(typ) is getattr(_typing, name, None):
                return origin

        return _get_origin(typ)

    _update_all("get_origin")


if "get_args" not in globals():
    from typing_inspect import get_args as _get_args  # type: ignore

    @_functools.wraps(_get_args)
    def get_args(typ):
        return _get_args(typ, True)

    _update_all("get_args")


# Build missing generic types.
_GenericInfo = _typing.NamedTuple(
    "_GenericInfo",
    (
        ("names", "_typing.Tuple[str, ...]"),
        ("args", "_typing.Tuple[_typing.TypeVar, ...]"),
    ),
)

_T = _typing.TypeVar("_T")
_KT = _typing.TypeVar("_KT")
_VT_co = _typing.TypeVar("_VT_co", covariant=True)
_T_co = _typing.TypeVar("_T_co", covariant=True)

_GENERIC_TYPES = {
    ReferenceType: _GenericInfo(
        names=("ReferenceType", "ref"),
        args=(_typing.cast(_typing.TypeVar, _T),),
    ),
    WeakSet: _GenericInfo(
        names=("WeakSet",),
        args=(_typing.cast(_typing.TypeVar, _T_co),),
    ),
    WeakKeyDictionary: _GenericInfo(
        names=("WeakKeyDictionary",),
        args=(_typing.cast(_typing.TypeVar, _KT), _typing.cast(_typing.TypeVar, _VT_co)),
    ),
    WeakValueDictionary: _GenericInfo(
        names=("WeakValueDictionary",),
        args=(_typing.cast(_typing.TypeVar, _KT), _typing.cast(_typing.TypeVar, _VT_co)),
    ),
}

for _base, (_names, _vars) in _GENERIC_TYPES.items():
    _update_all(*_names)

    if hasattr(_base, "__class_getitem__") or hasattr(type(_base), "__getitem__"):
        continue

    if hasattr(_types, "new_class"):
        _generic = _types.new_class(_base.__name__, (_base, _typing.Generic[_vars]))  # type: ignore
    elif hasattr(_typing, "GenericMeta"):
        _generic = _typing.GenericMeta(_base.__name__, (_base, _typing.Generic[_vars]), {})  # type: ignore
    else:
        _generic = type(_base.__name__, (_base, _typing.Generic[_vars]), {})  # type: ignore

    for _name in _names:
        globals()[_name] = _generic


# Function to get type name that supports generics.
_SPECIAL_NAMES = {
    Any: "Any",
    ClassVar: "ClassVar",  # noqa
    Optional: "Optional",
    Literal: "Literal",
    Final: "Final",
    Union: "Union",
    ForwardRef: "ForwardRef",
    Ellipsis: "...",
    None: "None",
    type(None): "NoneType",
    True: "True",
    False: "False",
    NotImplemented: "NotImplemented",
    type(NotImplemented): "NotImplementedType",
}


def get_name(typ, qualname_getter=lambda t: getattr(t, "__qualname__", None)):
    # type: (Any, Callable[[Any], Optional[str]]) -> Optional[str]
    """
    Get name.

    :param typ: Type/typing form.
    :param qualname_getter: Qualified name getter function override.
    :return: Name or None.
    """
    name = None

    # Special name.
    try:
        if typ in _SPECIAL_NAMES:
            name = _SPECIAL_NAMES[typ]
    except TypeError:  # ignore non-hashable
        pass

    # Python 2.7.
    if name is None:
        if not hasattr(typ, "__forward_arg__") and type(typ).__module__ in ("typing", "typing_extensions", "tippo"):
            if type(typ).__name__.lstrip("_") == "Literal":
                return "Literal"
            if type(typ).__name__.lstrip("_") == "Final":
                return "Final"
            if type(typ).__name__.lstrip("_") == "ClassVar":
                return "ClassVar"

    # Try a couple of other ways to get the name.
    if name is None:

        # Get origin name.
        origin = get_origin(typ)
        if origin is not None:
            origin_name = (
                qualname_getter(origin)
                or getattr(origin, "__name__", None)
                or getattr(origin, "_name", None)
                or getattr(origin, "__forward_arg__", None)
            )
        else:
            origin_name = None

        # Get the name.
        name = (
            qualname_getter(typ)
            or getattr(typ, "__name__", None)
            or getattr(typ, "_name", None)
            or getattr(typ, "__forward_arg__", None)
        )

        # We have an origin name.
        if origin_name:

            # But we don't have a name. Use the origin name instead.
            if not name:
                name = origin_name

            # Prefer the origin name if longer and (for qualified generic names).
            elif name and origin_name.endswith(name) and len(origin_name) > len(name):
                name = origin_name

        # Special cases.
        elif not name:
            for cls_name, special_name in {
                "_Literal": "Literal",
                "_ClassVar": "ClassVar",
                "_Final": "Final",
                "_Union": "Union",
            }.items():
                cls = getattr(_typing, cls_name, None)
                if type(typ) is cls or origin is not None and type(origin) is cls:
                    name = special_name
                    break

    return name or None


_update_all("get_name")


class SupportsKeysAndGetItem(Protocol[_KT, _VT_co]):
    """Protocol that can be used as an input to a dictionary constructor."""

    def keys(self):
        # type: () -> Iterable[_KT]
        """
        Get keys.

        :return: Keys.
        """
        pass

    def __getitem__(self, __k):
        # type: (_KT) -> _VT_co
        """
        Get value for key.

        :param __k: Key.
        :return: Value.
        """
        pass


_update_all("SupportsKeysAndGetItem")



# type: ignore

import os
import sys

# Path setup.
root_path = os.path.abspath(os.path.join("..", ".."))
sys.path.insert(0, root_path)

# Project information.
project = "Tippo"
copyright = "2022, Bruno Nicko"  # noqa
author = "Bruno Nicko"

# Sphinx extensions.
extensions = [
    "sphinx.ext.intersphinx",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.autodoc",
    "sphinx_rtd_theme",
]

# Intersphinx configuration.
intersphinx_mapping = {
    "six": ("https://six.readthedocs.io/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
    "python": ("https://docs.python.org/3.10", None),
}

# The suffix(es) of source filenames.
source_suffix = [".rst"]

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx.
language = "en"

# List of patterns to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "monokai"

# HTML options.
html_theme = "sphinx_rtd_theme"
html_theme_options = {
    "style_external_links": False,
    "style_nav_header_background": "#9966CC",
}
html_static_path = ["_static"]
